"""
Daves utilities is a collection of useful python functions which I would like to
provide to anybody who is interested in using them.
"""

__version__ = "0.0.14"

from daves_utilities.for_long import for_long
from daves_utilities.fun_save import fun_save
from daves_utilities.is_equal import is_equal
from daves_utilities.print_structure import print_str

__all__ = ["for_long", "fun_save", "is_equal", "print_str"]



import numpy as np
import pandas as pd


def is_equal_recursive(elem1, elem2):  # noqa

    if type(elem1) != type(elem2):
        return False
    if isinstance(elem1, list):
        if len(elem1) != len(elem2):
            return False
        return [is_equal_recursive(i, j) for i, j in zip(elem1, elem2)]
    elif isinstance(elem1, dict):
        if len(elem1) != len(elem2):
            return False
        elem1_in = list(elem1.values())
        elem2_in = list(elem2.values())
        return [is_equal_recursive(i, j) for i, j in zip(elem1_in, elem2_in)]
    elif isinstance(elem1, np.ndarray):
        return (elem1 == elem2).all()
    elif "pandas" in str(type(elem1)):
        if elem1.shape != elem2.shape:
            return False
        return elem1.equals(elem2)
    elif "sklearn" in str(type(elem1)):
        return str(elem1) == str(elem2)
    else:
        try:
            return float(elem1) == float(elem2)
        except Exception as e:
            print("No fitting type for the elements was found!")
            raise e


def list_flatten(l_in, a=None):
    if a is None:
        a = []
    l_in = [l_in] if isinstance(l_in, bool) else l_in
    for i in l_in:
        if isinstance(i, list):
            list_flatten(i, a)
        else:
            a.append(i)
    return a


def is_equal(elem1, elem2, flatten=True):
    boolen_list = is_equal_recursive(elem1, elem2)

    if flatten:
        return all(list_flatten(boolen_list))
    else:
        return boolen_list


if __name__ == "__main__":
    df2 = pd.DataFrame(
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), columns=["a", "b", "c"]
    )

    test1 = [{"a": 1, "b": 2}, 10, [1, 2, [4, 5, 6]], df2]
    test2 = [{"a": 1, "b": 2}, 10, [1, 2, [4, 5, 6]], df2]

    out = is_equal(test1, test2)
    print(out)



import itertools
import os
import re
import shutil
from os import listdir
from pathlib import Path

import pandas as pd
from progressbar import progressbar  # pip install progressbar2


def for_long(iter_fun, iter_attr, save_every_n=10, path="./"):  # noqa
    """
    for_long facilitates iterating over long running loops.

    following functionalities are implemented:
    - save progress in folder
    - start from last saved progress
    - run over combination of input lists
    - get show loading bar

    iter_fun:       function to iterate over
    iter_attr:      a dictionary of all attributes needed for iter_fun input.
                    => {"attribute name 1":attribute_values1,
                        "attribute name 2":attribute_values2, ...}
                    for the loop to create a permutation you can enter one or more
                        attributes as a list.
                    => {"attribute name 1":[1,2,3,4], "attribute name 2":
                        ["EU","CH","AM"], ...}
                    attributes with lists can be combined with attributes with a
                        single value.
    save_every_n:   save my progress every n iterations
    path:           save my files under this path
    """

    # extract combinations
    combination_values = {j: i for j, i in iter_attr.items() if isinstance(i, list)}
    static_values = {j: i for j, i in iter_attr.items() if not isinstance(i, list)}
    # create combination
    value_combination = list(itertools.product(*list(combination_values.values())))

    # create dict from all combinations to input into function
    iter_input = [dict(zip(combination_values.keys(), i)) for i in value_combination]

    # define path for backup files
    path_backup = f"{path}for_long_{iter_fun.__name__}"

    # check for precalcuated combinations
    end_file_path = f"{path}/{iter_fun.__name__}_end.csv"
    counter = 0
    if os.path.isfile(end_file_path):
        # load las calculated end file
        print(f"Finsihed file {end_file_path} has been found")
        df_out = pd.read_csv(end_file_path)
    elif os.path.exists(path_backup):
        # find file with highest iter backup number
        print(f"Started backup files {path_backup} have been found")
        for i in listdir(path_backup):
            m = re.search(fr"{iter_fun.__name__}_(\d*).csv", i)
            if m and int(m.group(1)) > counter:
                counter = int(m.group(1))
        df_out = pd.read_csv(f"{path_backup}/{iter_fun.__name__}_{str(counter)}.csv")
    else:
        # create new empty data frame
        print("No precalculated data has been found for " + str(iter_fun.__name__))
        df_out = pd.DataFrame(columns=[*iter_input[0], *static_values, "output"])

    [i.update(static_values) for i in iter_input]

    # remove entries from iter_input which have already been calculated
    df_in = pd.DataFrame(iter_input)
    if len(df_in.columns) == len(df_out.columns) - 1:
        df_limited = df_in[~df_in.isin(df_out)].dropna()
        iter_attributes = df_limited.to_dict("records")
    else:
        print(
            """Number of attributes for the functions have changed. Function will run
                again over all combinations and append results."""
        )

    def iter_function(iter_in, static_values, df_out, counter):
        # run function
        iter_out = iter_fun(**iter_in, **static_values)

        # append output
        df_out = df_out.append(
            {**iter_in, **static_values, "output": iter_out}, ignore_index=True
        )

        # save output every n iterations
        if counter % save_every_n == 0 and counter != 0:
            # create folder for backup storage of files
            Path(f"{path}for_long_{iter_fun.__name__}").mkdir(
                parents=True, exist_ok=True
            )
            # write backup file to directory
            df_out.to_csv(
                f"{path_backup}/{iter_fun.__name__}_{counter}.csv", index=False
            )

        counter += 1

        return df_out

    # iterate over function
    for iter_in in progressbar(iter_attributes):
        df_out = df_out.append(iter_function(iter_in, static_values, df_out, counter))

    # save last data frame END
    df_out.to_csv(f"{path}/{iter_fun.__name__}_end.csv", index=False)

    # remove backup during calculation
    try:
        shutil.rmtree(path_backup)
    except:  # noqa
        pass

    return df_out


if __name__ == "__main__":

    # create input
    embeddings = ["tf-idf", "word2vec", "doc2vec", "bert"]
    embeddings_dim = [10, 20, 40, 100, 500, 1000]
    range_attr = list(range(0, 10))
    # combination
    attribute_dict = {
        "embeddings": embeddings,
        "embeddings_dim": embeddings_dim,
        "verbose": True,
        "quick_calc": False,
        "range_attr": range_attr,
    }

    # function to apply
    import time

    def concatenate_input(
        embeddings, embeddings_dim, verbose, quick_calc, range_attr=0
    ):
        time.sleep(0.01)
        return f"{embeddings} {str(embeddings_dim)} {verbose} {quick_calc} {range_attr}"

    df_out = for_long(iter_fun=concatenate_input, iter_attr=attribute_dict)



import os
import pickle

from . import is_equal


def fun_save(
    fun_input, attr_input={}, path="./", identifier: str = "", only_save_one=False
):

    # create hidden folder under path to save function outputs
    fun_save_path = path + ".fun_save"
    if not os.path.exists(fun_save_path):
        os.makedirs(fun_save_path)

    # add underline to identifier if it exists
    identifier = "_" + identifier if identifier else identifier

    # structure to check pas calculations
    save_structure = f"{fun_input.__name__}{identifier}"
    files_found = [
        f"{fun_save_path}/{i}" for i in os.listdir(fun_save_path) if save_structure in i
    ]

    # Warning if multiple files found while only_save_one = True
    if len(files_found) > 1 and only_save_one:
        print(
            """only_save_one is True but multiple files were found. Only the last file
            will be overwritten if not output/input combination is found."""
        )

    file_id = 0
    # open all files and compare attributes, return if same input/output is found
    for iter_path in files_found:
        # read files
        with open(iter_path, "rb") as f:
            output = pickle.load(f)

        # check if input is the same
        if is_equal.is_equal(output["attr_input"], attr_input):
            print("Function file has been found and reused: " + save_structure)
            return output["fun_output"]

        # increament id if larger
        file_id_iter = int(iter_path.split("_")[-1].replace(".pkl", ""))
        file_id = file_id_iter if file_id_iter > file_id else file_id

    # increment file_id
    if not only_save_one:
        file_id += 1
    print(
        "Input/Output combination not found => recalculating function now: "
        + save_structure
    )

    # create full file name with identifier for multiple input variations
    save_name = f"{fun_save_path}/{save_structure}_{file_id}.pkl"

    # run function
    fun_output = fun_input(**attr_input)
    output = {"attr_input": attr_input, "fun_output": fun_output}

    # save function output
    with open(save_name, "wb") as f:
        pickle.dump(output, f, protocol=4)

    return output["fun_output"]


if __name__ == "__main__":

    input_values = {"embeddings": ["tf-idf", "word2vec", "bert"]}

    # function to apply
    import time

    def concatenate_input(embeddings):
        time.sleep(1)
        return f"{embeddings}"

    output = fun_save(
        fun_input=concatenate_input,
        attr_input=input_values,
        path="./daves_utilities/data/",
        identifier="",
    )

    print(output)



"""
print_structure prints out the structure of any arbitrary python object..
"""

__version__ = "0.0.3"

import numpy as np
import pandas as pd


def print_str(object_atr, print_elements=False, level=0):
    """
    print_str is the wrapper command which recoursively calles itself,
        as well as the print function.
    """

    if isinstance(object_atr, (str, int, float, bool, complex, bytes, memoryview)):
        if print_elements:
            print_level_type(object_atr, level)
        else:
            return
    elif isinstance(object_atr, (pd.core.frame.DataFrame, pd.core.series.Series)):
        print_level_type(object_atr, level)
        if isinstance(object_atr, pd.core.frame.DataFrame):
            if print_elements:
                [
                    print_level_type(i, level + 1, pandas_dtype=j)
                    for i, j in zip(object_atr.columns, object_atr.dtypes)
                ]
    elif isinstance(object_atr, (list, tuple, range, set, frozenset, bytearray)):
        print_level_type(object_atr, level)
        [print_str(i, print_elements, level + 1) for i in object_atr]
    elif isinstance(object_atr, dict):
        print_level_type(object_atr, level)
        elem1_in = list(object_atr.values())
        [print_str(i, print_elements, level + 1) for i in elem1_in]
    elif isinstance(object_atr, np.ndarray):
        print_level_type(object_atr, level)


def get_shape(input_elem) -> str:
    if isinstance(input_elem, (list, dict, tuple, set, frozenset, range, bytearray)):
        return "(" + str(len(input_elem)) + ")"
    elif isinstance(
        input_elem, (pd.core.frame.DataFrame, pd.core.series.Series, np.ndarray)
    ):
        return str(input_elem.shape)
    else:
        return ""


def print_level_type(input_elem, level, pandas_dtype=None):
    # get shape
    shape_out = get_shape(input_elem)

    # create indentation
    level_indent = ""
    if level > 0:
        level_indent = "  " * level
    level_indent = level_indent + "|___"
    type_string = str(type(input_elem)).split("'")[1]

    # if pandas columns are given
    if pandas_dtype:
        level = ""
        level_indent += "__"
        type_string = f"{input_elem} - {pandas_dtype}"

    # create print statement
    print_color = f"{level_indent}{level} [{type_string}] {shape_out}"

    print(print_color)


if __name__ == "__main__":
    """
    Testing the print_structure function

    execute test on command line with:
    > pytest test_print_structure.py
    """

    from ..print_structure import print_str  # noqa

    # pandas data
    df2 = pd.DataFrame(
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), columns=["a", "b", "c"]
    )
    data = np.array(["a", "b", "c", "d"])
    s = pd.Series(data)

    # numpy data
    np_array = np.array([[1, 2], [3, 4]])

    # complex numbers
    complex_n = 2 + 3j
    # boolean
    boolean_var = True
    # tuple
    tuple_var = (1, 3, 4)
    # set
    set_var = {1, 2, 4}
    # range
    range_var = range(0, 3)

    # bytes
    simple_bytes_string = b"test_test"
    string = "Pyth"
    bytes_arr = bytearray(string, "utf-8")
    mem_var = memoryview(simple_bytes_string)

    test1 = [
        {"a": 1, "b": 2},
        10,
        [1, 2, [4, 5, 6, [1, 3, ["string", "test", [1, [1, [1]]]]]]],
        df2,
        [s],
        np_array,
        complex_n,
        boolean_var,
        tuple_var,
        set_var,
        range_var,
        simple_bytes_string,
        bytes_arr,
        mem_var,
    ]

    print_str(object_atr=test1)
